/*-----------------------------------------------------------------------------
  Title         : mcs.c
  Language      : ANSI C
  Description   : Calculates Mean Character Support (MCS),
                  Mean Resampling Proport (MRP), and
                  their standard deviations (sCS and sRP).
                  Reads from stdin, writes to stdout.
  Usage         : mcs TREEFILENAME
  Usage (alt.)  : mcs
  Input         : PAUP* (Swofford, 2000) treefile generated by using
                  commands:
                  boot keepall=yes; [or jack keepall=yes;]
                  savetrees format=nexus file=TREEFILENAME
                  savebootp=brlens;
  Warning:      : Beta, use at your own risk.
  Reference     : http://www.ebc.uu.se/systzoo/staff/nylander.html
  Author        : J.A.A. Nylander
  Version       : 1.0b
  Latest Update : 04-10-2001
  ----------------------------------------------------------------------------- */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stddef.h>
#include <ctype.h>


/********** GLOBAL VARIABLES**********/

int nTax, nReps, maxLength;
double mcs, scs, meancs, mincs, maxcs;
double mrp, srp, meanrp, minrp, maxrp;
char *tree, *treeline;
char treeFileName[100]; 
const char rooted[]= "tree PAUP_1 = [&R] ";
const char unrooted[]= "tree PAUP_1 = [&U] ";
const char completed[]= "replicates completed";
const char aborted[]= "aborted after ";
FILE *treeFile=NULL;


/********** PROTOTYPES **********/

static void PrintUsage();
int GetMaxLength (FILE *treeFile);
int GetnReps (FILE *treeFile);
int TaxCount (char *tree);
double Correct (double value);
double LN (double value);
void MCS (char *tree);
void StdDev (char *tree);


/********** FUNCTIONS **********/

/**** main */
main(int argc, char *argv[])
{

  /* Open treefile */
  if (argc <= 1) {
    PrintUsage();    
/*++pac
    fprintf (stdout, "Enter name of tree file: ");
fflush(stdout);	//++pac
    scanf("%s", treeFileName);
    getchar();
    treeFile=fopen(treeFileName,"r");
    if (!treeFile) {
      fprintf (stdout, "\a\nERROR: Could not open file: %s\n", treeFileName);
      fprintf (stdout, "Check spelling and/or file location.\nQuitting mcs...\n\n");		
      getchar();
      exit(0);
    }
    else  printf ("\nOpened %s\n", treeFileName); 

*/
  }	
  else if (argc > 2) {
    fprintf (stdout, "\a\nERROR: Too many arguments\n\n");
    PrintUsage();
    fprintf (stdout, "Quitting mcs...\n\n");
    getchar();
    exit(0);
  }
  else {
    treeFile = fopen(argv[1], "r");
    if (!treeFile) {
      fprintf (stdout, "\a\nERROR: Could not open file: %s\n", argv[1]);
      fprintf (stdout, "\nCheck spelling and/or file location.\nQuitting mcs...\n\n");		
      getchar();
      exit(0);
    }
    else fprintf (stdout, "\nOpened %s\n", argv[1]);
  }


  /* Get length of longest line in inputfile and allocate memory for tree and treeline */
  maxLength = GetMaxLength(treeFile);
  rewind(treeFile);
	
  tree=(char*)malloc((size_t)(maxLength+2)*sizeof(char));
  if (!tree) {
    fprintf(stdout, "\a\nERROR: Could not allocate memory for tree!\n");
    fprintf(stdout, "Quitting mcs...\n\n");    
    getchar();
    exit(0);
  }
  treeline=(char*)malloc((size_t)(maxLength+2)*sizeof(char));
  if (!treeline) {
    fprintf(stdout, "\a\nERROR: Could not allocate memory for treeline!\n");
    fprintf(stdout, "Quitting mcs...\n\n");    
    getchar();
    exit(0);
  }
	

  /* Get tree description in to tree */
  while (fgets(treeline, maxLength+1, treeFile) != NULL) {
    if (strstr(treeline,unrooted) != NULL || strstr(treeline,rooted) != NULL)
      strcpy(tree,treeline);
  }
	
  if (strstr(tree,unrooted) == NULL && strstr(tree,rooted) == NULL) {
    fprintf(stdout, "\a\nERROR: Could not find tree description.\n");
    fprintf(stdout, "Check file format (Description has to start with:\n");
    fprintf(stdout, "'tree PAUP_1 = [&R]' or 'tree PAUP_1 = [&U]'.)\n\n");
    fprintf(stdout, "Quitting mcs...\n\n");
    getchar();
    exit(0);
  }
	
  rewind(treeFile);
	


  /* Get number of taxa */
  nTax = TaxCount(tree);


  /* Get nReps*/
  nReps = GetnReps(treeFile);


  /* Get MCS and MRP and their standard deviations */
  MCS(tree);


  /* Get sCP and sRP */
  StdDev(tree);
  

  /* Print results */
  fprintf(stdout, "\nNumber of terminals              : %d\n", nTax);
  fprintf(stdout, "Number of pseudoreplicates       : %d\n\n", nReps);
  fprintf(stdout, "Mean Character Support (MCS)     : %.3f\n", mcs);
  fprintf(stdout, "Standard deviation of CS (sCS)   : %.3f\n", scs);
  fprintf(stdout, "Range of CS values               : %.3f - %.3f\n\n", mincs, maxcs);
  fprintf(stdout, "Mean Resampling Proportion (MRP) : %.3f\n", mrp);
  fprintf(stdout, "Standard deviation of RP (sRP)   : %.3f\n", srp);
  fprintf(stdout, "Range of RP's                    : %.3f - %.3f\n\n\n", minrp, maxrp);


  /* Clean up */
  fclose (treeFile);
  free (treeline);
  free (tree);

  getchar();
  return 0;
}


/**** PrintUsage - Print usage to screen */
void PrintUsage()
{ 	
  fprintf(stdout, "\n-------------------------------------------------------------------\n");
  fprintf(stdout, "Usage:   mcs TREEFILENAME\n\n");
  fprintf(stdout, "(or type TREEFILENAME when asked for.)\n\n");
  fprintf(stdout, "'TREEFILENAME' above is the name of tree file in NEXUS format.\n");
  fprintf(stdout, "The tree file should have been generated in PAUP* (Swofford, 2000)\n");
  fprintf(stdout, "by parsing commands: boot keepall=yes; [or jack keepall=yes;]\n");
  fprintf(stdout, "savetrees format=nexus file=TREEFILENAME savebootp=brlens;\n");
  fprintf(stdout, "-------------------------------------------------------------------\n\n");
}


/**** TaxCount - Count Number of taxa */
int TaxCount (char *tree)
{
  int i, j=0;
  for (i=0; tree[i]; i++) {
    if (tree[i]==',')
      j++;
  }
  return (j+1);
}


/**** GetMaxLength - Get length of longest line in inputfile */
int GetMaxLength (FILE *treeFile)
{
  int length, maxL=0;
  char c=0;
  while (c!=EOF) {
    c=fgetc(treeFile);
    length=0;
    while (c!=EOF && c!='\n' && c!='\r') {
      length++;
      c=fgetc(treeFile);
    }
    if (length > maxL)
      maxL = length;
  }
  return(maxL);
}


/**** GetnReps - Get number of pseudoreplicates, nReps. */
int GetnReps (FILE *treeFile)
{
  int i, j;
  char nr[15];
  char *rep, *repline;
  const char boot[]= "ootstrap";
  const char jack[]= "ackknife";
	
  rep=(char*)malloc((size_t)(maxLength+2)*sizeof(char));
  if (!rep) {
    fprintf(stdout, "\a\nERROR: Could not allocate memory for rep!\n");
    exit(0);
  }

  repline=(char*)malloc((size_t)(maxLength+2)*sizeof(char));
  if (!repline) {
    fprintf(stdout, "\a\nERROR: Could not allocate memory for repline!\n");
    exit(0);
  }
	
  while (fgets(repline, maxLength+1, treeFile) != NULL) {
		
    if (strstr(repline,completed) != NULL || strstr(repline,aborted) != NULL)
      strcpy(rep,repline);
  }

  if (strstr(rep,boot) == NULL && strstr(rep,jack) == NULL) {
    fprintf(stdout, "\a\nERROR: Could not find number of replicates.\n");
    fprintf(stdout, "\nCheck file format.\nQuitting mcs...\n");
    exit(0);
  }
	

  for (i=j=0; rep[i]!='\0' && rep[i]!='\n' && rep[i]!='\r'; i++)
    if (isdigit(rep[i])!=NULL)
      nr[j++]=rep[i];
	
  nr[j] = '\0';
		
  free(repline);
  free(rep);
	
  return (atoi(nr));
}


/**** Correct - Correct RP's for limited number of pseudoreplicates */
double Correct (double value)
{
  return ((value*nReps)/(nReps+1));
}

/**** LN - log RP's */
double LN (double value)
{
  return (log(1/(1-value)));
}


/**** MCS - Calculate MCS and MRP, and get the range  */
void MCS (char *tree)
{
  int i, j;
  char  *p, *q;
  char cs[] ="),";
  char start[]="):";
  char value[15];
  double cssum=0.0, rpsum=0.0, tal=0.0;
  double a, b, rpmax=0.0, rpmin=1.0, csmax=0.0, csmin;
 
  j=0;
  p = strstr(tree, start) +1;
  csmin = log(1/(1-((nReps-0.5)/nReps)));
  
  while (p!=NULL) {
    q = strpbrk(p,cs);
		
    for (i=0 ; p<q ;p++) {
      if(isdigit(*p) || *p=='.')
	value[i++] = *p;
    }
    value[i]='\0';
		
    tal = Correct(atof(value))/100.0;
		
    rpsum += tal;
    cssum += LN(tal);
	
	/* Get the ranges */
	a = tal;
	if (a > rpmax)
		rpmax = tal;
	if (a < rpmin)
		rpmin = tal;	
	minrp = rpmin;
	maxrp = rpmax;
	
	b = LN(tal);
	if (b > csmax)
		csmax = LN(tal);
	if (b < csmin)
		csmin = LN(tal);	
	mincs = csmin;
	maxcs = csmax;

	
    j++;
		
    p = strstr(q, start);
		
    if (p==NULL)
      break;
    else p++;
  }
	
  mcs = (cssum/(nTax-3));
  mrp = (rpsum/(nTax-3));
  meanrp = rpsum/j;
  meancs = cssum/j;
	
  return;
}	


/**** StdDev - Calculate the standard deviations, sCS and sRP */
void StdDev (char *tree)
{
  int i, j;
  char  *p, *q;
  char cs[] ="),";
  char start[]="):";
  char value[15];
  double sqrsumRP=0, sqrsumCS=0, tal=0;

  p = strstr(tree, start) +1;
	
  j=0;
	
  while (p!=NULL) {
    q = strpbrk(p,cs);
		
    for (i=0 ; p<q ;p++) {
      if(isdigit(*p) || *p=='.')
	value[i++] = *p;
    }
    value[i]='\0';
		
    tal = Correct(atof(value))/100.0;
		
    sqrsumRP += (tal-meanrp)*(tal-meanrp);
    sqrsumCS += (LN(tal)-meancs)*(LN(tal)-meancs);
		
    j++;
		
    p = strstr(q, start);
		
    if (p==NULL)
      break;
    else p++;
  }

  srp = sqrt(sqrsumRP/(j-1));
  scs = sqrt(sqrsumCS/(j-1));

  return;
}	

